From e04cd6f51e2745337c57b6a2fabe0c56d8be4aa1 Mon Sep 17 00:00:00 2001
From: Feng Xue <fxue@os.amperecomputing.com>
Date: Fri, 3 Jan 2020 22:00:55 +0800
Subject: [PATCH] Fix pr93084

---
 gcc/ipa-cp.c                           | 21 +++++++++++--
 gcc/testsuite/gcc.dg/ipa/ipa-clone-3.c | 42 ++++++++++++++++++++++++++
 2 files changed, 61 insertions(+), 2 deletions(-)
 create mode 100644 gcc/testsuite/gcc.dg/ipa/ipa-clone-3.c

diff --git a/gcc/ipa-cp.c b/gcc/ipa-cp.c
index 4381b35a809..11631986293 100644
--- a/gcc/ipa-cp.c
+++ b/gcc/ipa-cp.c
@@ -1917,10 +1917,25 @@ self_recursively_generated_p (ipcp_value<tree> *val)
 
       class ipcp_param_lattices *plats = ipa_get_parm_lattices (info,
 								src->index);
-      ipcp_lattice<tree> *src_lat = src->offset == -1 ? &plats->itself
-						      : plats->aggs;
+      ipcp_lattice<tree> *src_lat;
       ipcp_value<tree> *src_val;
 
+      if (src->offset == -1)
+	src_lat = &plats->itself;
+      else
+	{
+	  struct ipcp_agg_lattice *src_aglat;
+
+	  for (src_aglat = plats->aggs; src_aglat; src_aglat = src_aglat->next)
+	    if (src_aglat->offset == src->offset)
+	      break;
+
+	  if (!src_aglat)
+	    return false;
+
+	  src_lat = src_aglat;
+	}
+
       for (src_val = src_lat->values; src_val; src_val = src_val->next)
 	if (src_val == val)
 	  break;
@@ -2017,6 +2032,8 @@ propagate_vals_across_arith_jfunc (cgraph_edge *cs,
 	    val_seeds.safe_push (src_val);
 	}
 
+      gcc_assert ((int) val_seeds.length () <= param_ipa_cp_value_list_size);
+
       /* Recursively generate lattice values with a limited count.  */
       FOR_EACH_VEC_ELT (val_seeds, i, src_val)
 	{
diff --git a/gcc/testsuite/gcc.dg/ipa/ipa-clone-3.c b/gcc/testsuite/gcc.dg/ipa/ipa-clone-3.c
new file mode 100644
index 00000000000..18d29bdd0b3
--- /dev/null
+++ b/gcc/testsuite/gcc.dg/ipa/ipa-clone-3.c
@@ -0,0 +1,42 @@
+/* { dg-do compile } */
+/* { dg-options "-O3 -fdump-ipa-cp-details -fno-early-inlining --param ipa-cp-max-recursive-depth=8 --param ipa-cp-eval-threshold=1" } */
+
+struct V {
+  int f0;
+  int f1;
+};
+
+int data[100];
+
+int fn ();
+
+int recur_fn (struct V * __restrict v)
+{
+  int i = v->f0;
+  int j = v->f1;
+  struct V t;
+
+  if (j > 100)
+    {
+      fn ();
+      return 1;
+    }
+
+  data[i] = i;
+
+  t.f0 = i - 2;
+  t.f1 = j + 1;
+
+  recur_fn (&t);
+
+  return i * j;
+}
+
+int main ()
+{
+  struct V v = {1, 3};
+
+  return recur_fn (&v);
+}
+
+/* { dg-final { scan-ipa-dump-times "Creating a specialized node of recur_fn/\[0-9\]*\\." 8 "cp" } } */
-- 
2.17.1

